#include "stock.h"
#include "threads.h"

static pthread_mutex_t player_tot_mut = PTHREAD_MUTEX_INITIALIZER;
plid_t player_tot;
plid_t player_simult_tot;

int server2;

static void add_new_player (int client, plid_t plid,
			    passwd_t passwd, pthread_t *th);
static void validate_player_reentry (int client, passwd_t passwd);
inline static void kick_player_during_login (int client);

plid_t do_login (int client, pthread_t *th) {
   passwd_t passwd;
   req_t req;
   plid_t plid=-1;
   int id='s';  /* thread identifier : s - sync channel, a - async channel */
   
   sr_req (req);
   sr_passwd(passwd);
   
#ifdef DEBUG
   printf ("req : %d\t passwd : %d\n", req, passwd);
#endif  
   
   switch (req) {
      
    case NEW_LOGIN:                    /* New Player */
      if ((player_simult_tot == max_simult_players) ||
	  (player_tot == max_players)) {
	 
#ifdef DEBUG
	 printf ("Server has reached limit of new connections\n");
#endif
	   
	 kick_player_during_login (client);
      }
	       
      pthread_mutex_lock(&player_tot_mut);
      plid = ++player_tot;       /* CR for player_tot */
      ++player_simult_tot;
      pthread_mutex_unlock(&player_tot_mut);
                        
      pthread_setspecific (plid_TSD, (void *) plid);

      add_new_player (client, plid, passwd, th);
      SEND_OK;

      pthread_setspecific (id_TSD, (void *) id);

      return plid;
      break;

    case RE_ENTRY:                   /* reentry into game */
      validate_player_reentry (client, passwd);
      pthread_setspecific (plid_TSD, (void *) plid);
      pthread_setspecific (id_TSD, (void *) id);
      SEND_OK;
      return plid;
      break;
      
    default:
#ifdef LOG
      fprintf (logfile, "plid %d : bad login\n", plid);
#endif
      kick_player_during_login (client);
   }
   return -1;         /* control shouldn't reach here though 
		       * or else the universe would be inconsistent */
}

void do_login2 (plid_t plid) {
   int client;
   req_t req;
   passwd_t passwd;
   thread_arg_t *thread_arg;
   player_t *p = &player[plid];
   pthread_t *async_th;
   
   do {
      if ((client = accept (server2, NULL, NULL)) < 0) {
	 p->client2 = p->client;
#ifdef DEBUG
	 perror ("accept");
#endif
	 kick_player ();
      }
      
      sr_req (req);
      sr_passwd (passwd);

      if ((passwd != p->passwd) || (req != NEW_LOGIN2)) {
#ifdef DEBUG
	 printf ("plid %d : Trouble on port 2\n", plid);
#endif
	 SEND_ERR;
	 close (client);
      }
   } while (passwd != p->passwd);
   
   p->client2 = client;

   thread_arg = (thread_arg_t *) xmalloc (sizeof(thread_arg_t));
   async_th = (pthread_t *) xmalloc (sizeof(pthread_t));
   thread_arg->client = client;
   thread_arg->plid = plid;
   thread_arg->th = async_th;
   
   if (pthread_create (async_th, &attr_detached,
			serve_requests_async, (void *) thread_arg) != 0) {
#ifdef DEBUG
      printf ("Error creating second thread for plid : %d\n", plid);
#endif
      /* This is a special error handler because kick_player() can't be called
       * at this point.
       */
      close (p->client);
      
      pthread_mutex_lock(&player_tot_mut);
      --player_tot;       /* CR for player_tot */
      --player_simult_tot;
      pthread_mutex_unlock(&player_tot_mut);
      
      kick_player_during_login (client);
      return;
   }
   
   setup_result_list (plid);
   SEND_OK;
   
#ifdef LOG
   fprintf (logfile, "new player : %d\n", player_tot);
#endif
   return;
}

static void add_new_player (int client, plid_t plid,
			    passwd_t passwd, pthread_t *th) {

   register int i;
   short int len;
   player_t *p;
   
   if (plid % resize == 0)
     player = (player_t *) xrealloc (player, 
				     (plid + resize) * sizeof (player_t));
   
   p = &player[plid];   
   
   p->client = client;
   p->passwd = passwd;
   p->cash = initial_cash;
   p->sync_th = th;
   p->async_th = th;
   p->timeout = timeout;
   
#ifdef LIFE
   pthread_mutex_init (&p->player_mut, NULL);
#endif
   
   pthread_mutex_init (&p->result_mut, NULL);
   pthread_cond_init (&p->result_cond, NULL);
   
   p->holding = (stock_player_t **) xmalloc (hashsize *
					     sizeof(stock_player_t *));
   
   sr_s_int (client, len);

   if (len > max_len_player_name) {
      kick_player ();
   }
   
   p->name = (char *) xmalloc (len + 1);
   sr (client, p->name, len);
   sw_plid (plid);

   p->holding = (stock_player_t **) xmalloc (hashsize * 
					     sizeof(stock_player_t *));
   
   for (i=0; i < hashsize; ++i)
     p->holding[i] = NULL;
   
   return;
}

static void validate_player_reentry (int client, passwd_t passwd) {
   plid_t plid=0;
   player_t *p;

   sr (client, &plid, sizeof(plid_t));
   
   p = &player[plid];
   
   if (p->passwd == passwd && p->client == -1) {
      p->client = client;
      return;
   } else kick_player_during_login (client);
   return;
}

inline static void kick_player_during_login (int client) {
   /* This version of kick_player () is called when only one thread
    * is present for the player and the login process is not 
    * complete.
    */

   close (client);   
   pthread_exit (NULL);
   return;
}

void kick_player () {
   /* The way the threads are terminated is not as good as good can be...
    * this has to be worked out better.. 
    */

   plid_t plid;
   int id;
   player_t *p;
   void *t;

   if ((t = pthread_getspecific (plid_TSD)) == NULL) {
#ifdef DEBUG
      printf ("plid not assigned yet\n");
#endif
      pthread_exit (NULL);      
   }
   plid = (int) t;
   
   if ((t = pthread_getspecific (id_TSD)) == NULL) {
#ifdef DEBUG
      printf ("id not assigned yet\n");
#endif
      pthread_exit (NULL);
   }
   id = (int) t;
   
   p = &player[plid];
   
#ifdef LOG
   fprintf (logfile, "kicking plid %d\n", plid);
#endif
   
   if (id == 's') {
      close (p->client2);
      pthread_cancel (*p->async_th);
      pthread_exit (NULL);
   } else if (id == 'a') {
      close (p->client);
      pthread_cancel (*p->sync_th);
      pthread_exit (NULL);
   }
#ifdef DEBUG
   else {
      printf ("Some screwup with thread keys\n");
   }
#endif 
   
   return;
}
